package it.univaq.di.choreos.elts.strongsimulation;

import it.univaq.di.choreos.ELTS.InitialState2State;
import it.univaq.di.choreos.ELTS.LabeledState2State;
import it.univaq.di.choreos.ELTS.State;
import it.univaq.di.choreos.ELTS.State2State;
import it.univaq.di.choreos.elts.strongsimulation.impl.DummyResult;

import java.util.ArrayList;
import java.util.List;

/**
 * 
 * @author Alexander Perucci
 *
 */
public class StrongSimulation {
	//contains the pair states analyzed, the object is the same for each recursive call 
	private List<StatesPairs> statesPairs;
	private Result result;
	private ELTSModel firstModelManager;
	private ELTSModel secondModelManager;



	public StrongSimulation(ELTSModel firstModelManager, ELTSModel secondModelManager ) {
		this.firstModelManager= firstModelManager;
		this.secondModelManager= secondModelManager;
		this.result= new DummyResult();
	}

	/**
	 * 
	 * @param result result to set
	 */
	public StrongSimulation(ELTSModel firstModelManager, ELTSModel secondModelManager, Result result){
		//create list of State Pair
		this.firstModelManager= firstModelManager;
		this.secondModelManager= secondModelManager;
		this.result = result;
	}


	/**
	 * 
	 * @param firstModelManager first model manager
	 * @param secondModelManager second model manager
	 * @return a value that indicates if first model can be strongly simulated by second model
	 */
	public boolean checkStrongSimulation(){
		//create list of State Pair
		this.statesPairs=new ArrayList<StatesPairs>();
		//normalize first model
		firstModelManager.getModelNormalized();
		//normalize second model
		secondModelManager.getModelNormalized();


		//get all transition from initial state first model
		List<InitialState2State> transitionFromInitiaStateFirstModel=firstModelManager.getAllInitial2StateTransitions();
		//get all transition from initial state second model
		List<InitialState2State> transitionFromInitiaStateSecondModel=secondModelManager.getAllInitial2StateTransitions();

		//check if that for each intial2State transitions of the first model there is a transition in the second model
		for (InitialState2State initial2State : transitionFromInitiaStateFirstModel) {
			if (!checkStrongSimulationFromInitial2State(initial2State, transitionFromInitiaStateSecondModel))
				return false;
		}

		/*
		 * if they are here means that all the initial2State transition in the first model are simulated in the second model,
		 * then the first model can then be simulated by the second model
		 */
		return true;

	}


	/**
	 * 
	 * @param initial2StateFirstModel initial to state transition in first model
	 * @param initial2StatesSecondModel list all initial to state transitions in second model
	 * @return a value that indicates if the initial to state transition in first model, can be strongly simulated by at least one transition in the second model
	 */
	private boolean checkStrongSimulationFromInitial2State(InitialState2State initial2StateFirstModel, List<InitialState2State> initial2StatesSecondModel){
		//get the target state from the initial2State transition in first model
		State stateFirstModel =initial2StateFirstModel.getTarget();

		//check if there is at least a strong simulation
		for (InitialState2State initial2StateSecondModel : initial2StatesSecondModel) {
			//get the target state from the initial2State transition selected in second model
			State stateSecondModel = initial2StateSecondModel.getTarget();
			if (checkStrongSimulationFromStates(stateFirstModel, stateSecondModel))
				return true;
		}

		/*
		 * if they are here means that the initial2State transition of first model can not be strongly simulated by any transition in the second model
		 */
		return false;
	}



	/**
	 * 
	 * @param firstState first state in first model
	 * @param secondState second state in second model
	 * @return a value that indicates if the first state can be strongly simulated by the second state
	 */
	private boolean checkStrongSimulationFromStates(State firstState, State secondState){
		//create the pair of states
		StatesPairs statesPair = new StatesPairs(firstState,secondState);
		result.writeMessage(statesPair.getFirstState().getName()+" ? "+statesPair.getSecondState().getName());

		//check if the current pair of states had already been analyzed, if so I can finish correctly
		if (!statesPairs.contains(statesPair)) 
			statesPairs.add(statesPair);
		else{
			result.writeMessage("the pair ("+firstState.getName()+","+secondState.getName()+ ") has already been analyzed");
			return true;
		}

		//gel all transitions from the first state
		List<LabeledState2State> allTransitionFirstState = firstModelManager.getAllState2StateTransitionsFromStateSource(firstState);
		//get all transitions from the second state
		//List<State2StateTransition> allTransitionSecondState = modelManager.getAllState2StateTransitionsFromStateSource(secondModel, secondState);

		//CONDITION TERMINATIONS RECURSION

		/*
		 * We consider automata in which every state is an accepting state
		 * (i.e., we consider LTSs). A consequence is that if an automaton
		 * accepts a particular component interaction seen as a sequence of
		 * component interface operation invocations, then it also accepts any
		 * initial part of that interaction/sequence. In other words, due to the
		 * finiteness of the set of component interface operations, although all
		 * the possible component interactions can be infinite we can always
		 * finitely represent them since the language built over the component
		 * interface operations (i.e., the model of the component interaction
		 * behavior) is prefix-closed. Prefix-closed languages are generated by
		 * prefix-grammars that describe exactly all regular languages.
		 * Moreover, we consider LTSs with only one Initial state and only one
		 * Final state that are used for gluing one LTS with another one.
		 */
		if (firstModelManager.existTransitionToFinalState(firstState) && !checkFinalStateCondition(secondState,new ArrayList<State>())){
			result.writeMessage(firstState.getName()+" has a transition to the final state and "+secondState.getName()+ " has not a transition to the final state");
			return false;
		}

		// each transitions of the first state must be simulated by at least one transition of the second state
		for (LabeledState2State state2StateTransition : allTransitionFirstState) {
			//get the allowed transitions from the second state (the allowed transition are transitions that have the same name of the transition to be analyzed )
			List<LabeledState2State> transitionAllowedToSecondState =secondModelManager.getAllState2StateTransitionsAllowable(state2StateTransition,secondState);
			//if there aren't allowed transitions means that the first state can not be simulated by the second state
			if (transitionAllowedToSecondState.size()==0)
				return false;
			else{
				//this variable is used to understand the reason for the termination of the cycle for
				boolean transitionOK=false;
				//check if there is at least one transition able to simulate the selected transition
				for (LabeledState2State state2StateTransitionAllowed : transitionAllowedToSecondState) {
					if (checkStrongSimulationFromStates (state2StateTransition.getTarget(), state2StateTransitionAllowed.getTarget())){
						transitionOK =true;
						break;
					}
				}

				if (!transitionOK){
					result.writeMessage(firstState.getName()+" can not be simulated by "+secondState.getName());
					return false;
				}
			}
		}


		//get all epsilon transition from first state
		for (State2State epsilonState2State : firstModelManager.getAllEpsilonState2StateTransitionsFromStateSource(firstState)) {
			if (!checkStrongSimulationFromStates (epsilonState2State.getTarget(), secondState)){
				return false;
			}

		}

		return true;

	}




	/**
	 * 
	 * @param secondState state
	 * @param states store the state analyzed
	 * @return a value that indicates if the state has state2Final transition
	 */
	private boolean checkFinalStateCondition (State secondState, List<State> states){
		//check if state has already been analyzed
		if (!states.contains(secondState)) 
			states.add(secondState);
		else
			return false;
		// check if state state has a transition to final state
		if (secondModelManager.existTransitionToFinalState( secondState)){
			return true;
		}
		// get all state2state transitions (epsilon) 
		List<State2State> allEpsilonState2StateTransitionSecondState =secondModelManager.getAllEpsilonState2StateTransitionsFromStateSource(secondState);
		for (State2State epsilonState2State : allEpsilonState2StateTransitionSecondState) {
			if (checkFinalStateCondition(epsilonState2State.getTarget(), states))
				return true;
		}

		return false;

	}



}
