package it.univaq.di.choreos.synthesis.synthesisprocessor.impl;

import it.univaq.di.choreos.synthesis.behavioursimulator.BehaviourSimulator;
import it.univaq.di.choreos.synthesis.behavioursimulator.BehaviourSimulatorException;
import it.univaq.di.choreos.synthesis.cdgenerator.CDGenerator;
import it.univaq.di.choreos.synthesis.cdgenerator.CDGeneratorException;
import it.univaq.di.choreos.synthesis.datamodel.CD;
import it.univaq.di.choreos.synthesis.datamodel.CLTSModel;
import it.univaq.di.choreos.synthesis.datamodel.ChoreographyBPMNSpecification;
import it.univaq.di.choreos.synthesis.datamodel.ELTSModel;
import it.univaq.di.choreos.synthesis.datamodel.Participant;
import it.univaq.di.choreos.synthesis.datamodel.Service;
import it.univaq.di.choreos.synthesis.datamodel.ServiceDiscoveryInput;
import it.univaq.di.choreos.synthesis.datamodel.SynthesisResult;
import it.univaq.di.choreos.synthesis.m2mtransformator.M2MTransformator;
import it.univaq.di.choreos.synthesis.m2mtransformator.M2MTransformatorException;
import it.univaq.di.choreos.synthesis.servicediscovery.ServiceDiscovery;
import it.univaq.di.choreos.synthesis.servicediscovery.ServiceDiscoveryException;
import it.univaq.di.choreos.synthesis.synthesisprocessor.SynthesisProcessor;
import it.univaq.di.choreos.synthesis.synthesisprocessor.SynthesisProcessorException;

import java.io.StringWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.PropertyException;

import org.ow2.choreos.chors.ChoreographyDeployer;
import org.ow2.choreos.chors.ChoreographyNotFoundException;
import org.ow2.choreos.chors.EnactmentException;
import org.ow2.choreos.chors.datamodel.Choreography;
import org.ow2.choreos.chors.datamodel.ChoreographySpec;
import org.ow2.choreos.chors.datamodel.LegacyServiceSpec;
import org.ow2.choreos.services.datamodel.DeployableServiceSpec;
import org.ow2.choreos.services.datamodel.PackageType;
import org.ow2.choreos.services.datamodel.ServiceDependency;
import org.ow2.choreos.services.datamodel.ServiceSpec;
import org.ow2.choreos.services.datamodel.ServiceType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class SynthesisProcessorImpl implements SynthesisProcessor {

	private static Logger logger = LoggerFactory.getLogger(SynthesisProcessorImpl.class);

	private BehaviourSimulator behaviourSimulator;
	private CDGenerator cdGenerator;
	private M2MTransformator m2mTransformator;
	private ServiceDiscovery serviceDiscovery;
	private ChoreographyDeployer choreographyDeployer;

	public SynthesisProcessorImpl(BehaviourSimulator behaviourSimulator, CDGenerator cdGenerator, M2MTransformator m2mTransformator, ServiceDiscovery serviceDiscovery, ChoreographyDeployer choreographyDeployer) {
		this.behaviourSimulator = behaviourSimulator;
		this.cdGenerator = cdGenerator;
		this.m2mTransformator = m2mTransformator;
		this.serviceDiscovery = serviceDiscovery;
		this.choreographyDeployer = choreographyDeployer;
	}

	public SynthesisResult startSynthesis(ChoreographyBPMNSpecification bpmnSpecification) throws SynthesisProcessorException {
		SynthesisResult synthesisResult = new SynthesisResult();
		try {

			logger.info("start Synthesis");
			logger.info("call bpmn2clts");
			CLTSModel cltsModel = m2mTransformator.bpmn2clts(bpmnSpecification);
			logger.info("clts to coord");
			List<ParticipantWithDependencies> participantsWithDependencies = initParticipantsWithDependencies(m2mTransformator.clts2coords(cltsModel));

			for (ParticipantWithDependencies participant : participantsWithDependencies) {
				logger.info("call projection for participant=" + participant.getName());
				ELTSModel participantELTS = behaviourSimulator.projection(cltsModel, participant.getName());

				logger.info("call discoverServices for participant=" + participant.getName());
				// get operation input/output messages to put at the service
				// base
				ServiceDiscoveryInput serviceDiscoveryInput = new ServiceDiscoveryInput(bpmnSpecification.getChoreographyName(), participant.getName(), participantELTS);
				List<Service> matchedServices = serviceDiscovery.discoverServices(serviceDiscoveryInput);

				// simulateAndSelectOneService
				boolean result = false;
				Service foundService = null;
				for (Service service : matchedServices) {
					logger.info("call simulation form participant=" + participant.getName() + "and for selected service=" + service.getWsdl());
					if (result = behaviourSimulator.simulation(participantELTS, service.getELTS())) {
						logger.info("service found for participant= " + participant.getName());
						foundService = service;
						break;
					}
				}
				if (!result) {
					logger.info("no service found for the participant= " + participant.getName());
					synthesisResult.setResult(false);
					return synthesisResult;
				}

				participant.setService(foundService);
			}

			logger.info("generate CDS");
			generateCDs(bpmnSpecification.getChoreographyName(), cltsModel, participantsWithDependencies);

			logger.info("createChorSpec");
			ChoreographySpec chorSpec = createChorSpec(participantsWithDependencies);
			String chorID = choreographyDeployer.createChoreography(chorSpec);
			logger.info("choreographyDeployer.createChoreography=" + chorID);
			logger.info("choreographyDeployer.enact(chorID)");
			Choreography choreography = choreographyDeployer.enactChoreography(chorID);
			String chorString = choreography2String(choreography);
			if (!"".equals(chorString)) {
				synthesisResult.setResult(true);
				synthesisResult.setChorSpec(chorString);
			} else {
				synthesisResult.setResult(false);
			}
			logger.info("end Synthesis");
			return synthesisResult;

		} catch (M2MTransformatorException e) {
			logger.error("", e);
			throw new SynthesisProcessorException(e);
		} catch (BehaviourSimulatorException e) {
			logger.error("", e);
			throw new SynthesisProcessorException(e);
		} catch (ServiceDiscoveryException e) {
			logger.error("", e);
			throw new SynthesisProcessorException(e);
		} catch (EnactmentException e) {
			logger.error("", e);
			throw new SynthesisProcessorException(e);
		} catch (ChoreographyNotFoundException e) {
			logger.error("", e);
			throw new SynthesisProcessorException(e);
		} catch (CDGeneratorException e) {
			logger.error("", e);
			throw new SynthesisProcessorException(e);
		}
	}

	private ChoreographySpec createChorSpec(List<ParticipantWithDependencies> participants) {
		ChoreographySpec chorSpec = new ChoreographySpec();
		for (ParticipantWithDependencies participant : participants) {
			// Create the serviceSpec for the service that plays the role of the
			// related participant
			// ChoreographyServiceSpec serviceChoreographyServiceSpec = new
			// ChoreographyServiceSpec();
			// serviceChoreographyServiceSpec.setName(participant.getName());
			// serviceChoreographyServiceSpec.addRole(participant.getName());

			ServiceSpec serviceSpec= null;
			if (participant.getService().isLegacy()) {
				serviceSpec = new LegacyServiceSpec();
				serviceSpec.setName(participant.getName());
				List<String> nativeURIs = new ArrayList<String>();
				nativeURIs.add(participant.getService().getServiceURI());
				((LegacyServiceSpec)serviceSpec).setNativeURIs(nativeURIs);
				serviceSpec.setServiceType(ServiceType.SOAP);
				List<String> roles = new ArrayList<String>();
				roles.add(participant.getName());
				serviceSpec.setRoles(roles);
				// serviceChoreographyServiceSpec.setServiceSpec(serviceSpec);
				// add the Service Choreography Service Spec to chorSpec
				// chorSpec.addChoreographyServiceSpec(serviceChoreographyServiceSpec);
				chorSpec.addServiceSpec((LegacyServiceSpec)serviceSpec);
			}else{
				serviceSpec = new DeployableServiceSpec();
				serviceSpec.setName(participant.getName());
				((DeployableServiceSpec)serviceSpec).setEndpointName(participant.getName());
				((DeployableServiceSpec)serviceSpec).setPackageUri(participant.getService().getServiceURI());
				((DeployableServiceSpec)serviceSpec).setPackageType(PackageType.COMMAND_LINE);
				serviceSpec.setServiceType(ServiceType.SOAP);
				((DeployableServiceSpec)serviceSpec).setPort(participant.getService().getPort());
				List<String> roles = new ArrayList<String>();
				roles.add(participant.getName());
				serviceSpec.setRoles(roles);
				chorSpec.addServiceSpec((DeployableServiceSpec)serviceSpec);
			}
				// Create the dependencies
				for (CDData cddata : participant.getDependencies()) {
					// Creates the CD Chor Service Spec
					// ChoreographyServiceSpec cdChoreographyServiceSpec = new
					// ChoreographyServiceSpec();
					// cdChoreographyServiceSpec.setName(cddata.getName());
					// cdChoreographyServiceSpec.addRole(cddata.getParticipantTarget());
					DeployableServiceSpec cdServiceSpec = new DeployableServiceSpec();
					cdServiceSpec.setName(cddata.getName());
					cdServiceSpec.setEndpointName(cddata.getName());
					cdServiceSpec.setPackageUri(cddata.getUrl());
					cdServiceSpec.setPackageType(PackageType.EASY_ESB);
					cdServiceSpec.setServiceType(ServiceType.SOAP);
					List<String> cdRoles = new ArrayList<String>();
					cdRoles.add(participant.getName());
					cdServiceSpec.setRoles(cdRoles);
					// cdChoreographyServiceSpec.setServiceSpec(cdServiceSpec);
					// Creates the CD dependency respects to the service
					// proxified
					// ChoreographyServiceDependency
					// cdChoreographyServiceDependency
					// = new
					// ChoreographyServiceDependency(cddata.getParticipantTarget(),
					// cddata.getParticipantTarget());
					ServiceDependency cdChoreographyServiceDependency = new ServiceDependency(cddata.getParticipantTarget(), cddata.getParticipantTarget());
					cdServiceSpec.addDependency(cdChoreographyServiceDependency);
					// cdChoreographyServiceSpec.addDependency(cdChoreographyServiceDependency);

					// add the CD Service Choreography Service Spec to chorSpec
					// chorSpec.addChoreographyServiceSpec(cdChoreographyServiceSpec);
					chorSpec.addServiceSpec(cdServiceSpec);

					// Creates the dependency respects to CD
					// ChoreographyServiceDependency
					// serviceChoreographyServiceDependency = new
					// ChoreographyServiceDependency(cddata.getName(),
					// cddata.getParticipantTarget());
					ServiceDependency serviceDependency = new ServiceDependency(cddata.getName(), cddata.getParticipantTarget());
					// serviceChoreographyServiceSpec.addDependency(serviceChoreographyServiceDependency);
					serviceSpec.addDependency(serviceDependency);

				
				}

		}

		return chorSpec;
	}

	private void generateCDs(String choreographyName, CLTSModel cltsModel, List<ParticipantWithDependencies> participantsWithDependencies) throws CDGeneratorException {
		Map<String, Service> services = initServices(participantsWithDependencies);
		for (ParticipantWithDependencies participantWithDependencies : participantsWithDependencies) {
			for (CDData cdData : participantWithDependencies.getDependencies()) {
				// get service dependency
				String wsdlTargetParticipant = services.get(cdData.getParticipantTarget()).getWsdl();
				CD cd = new CD(choreographyName, cdData.getName(), wsdlTargetParticipant, cdData.getCoordModel());
				String urlCD = cdGenerator.generate(cd);
				cdData.setUrl(urlCD);
			}
		}

	}

	private Map<String, Service> initServices(List<ParticipantWithDependencies> participantsWithDependencies) {
		Map<String, Service> services = new HashMap<String, Service>();
		for (ParticipantWithDependencies participantWithDependencies : participantsWithDependencies) {
			services.put(participantWithDependencies.getName(), participantWithDependencies.getService());
		}

		return services;
	}

	private List<ParticipantWithDependencies> initParticipantsWithDependencies(List<Participant> participants) {
		List<ParticipantWithDependencies> result = new ArrayList<ParticipantWithDependencies>();
		for (Participant participant : participants) {
			ParticipantWithDependencies participantWithDependencies = new ParticipantWithDependencies(participant);
			result.add(participantWithDependencies);
		}
		return result;
	}

	public String choreography2String(Choreography choreography) {

		try {
			JAXBContext jaxbContext = JAXBContext.newInstance(Choreography.class);
			Marshaller marshaller = jaxbContext.createMarshaller();
			marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
			StringWriter stringWriter = new StringWriter();
			marshaller.marshal(choreography, stringWriter);
			return stringWriter.toString();
		} catch (PropertyException e) {
			logger.error("", e);
		} catch (JAXBException e) {
			logger.error("", e);
		}
		return "";
	}
}
